import click
import spglib  # We need to import spglib here to suppress the warnings


@click.command()
@click.option(
    "--filein",
    type=click.Path(exists=True),
    default="geometry.in",
    show_default=True,
    help="File name of input structure.",
)
@click.option(
    "--format",
    type=str,
    default="aims",
    show_default=True,
    help="Format of input structure.",
)
@click.option(
    "--fmax",
    type=float,
    default=0.01,
    show_default=True,
    help="Maximum residual force.",
)
@click.option(
    "--fix-symmetry",
    is_flag=True,
    help="Fix spacegroup symmetry for relaxation.",
)
@click.option(
    "--relax-unit-cell",
    is_flag=True,
    help="Also relax the lattice vectors.",
)
@click.option(
    "--trajectory",
    type=str,
    default="bfgs.traj",
    show_default=True,
    help="Name of the ASE trajectory file.",
)
def clims_prerelax(filein, format, fmax, fix_symmetry, relax_unit_cell, trajectory):
    """
    Relaxation of input structure FILEIN with the mace-mp-0 force field up to residual
    forces of FMAX. Here, only the atomic positions are relaxed, while the lattice
    vectors are kept fixed. The force field should work well for most non-magnetic bulk
    systems.
    """
    import warnings

    warnings.filterwarnings("ignore")

    from ase.optimize import BFGS
    from ase.io import read
    from ase.constraints import FixSymmetry, UnitCellFilter
    from clims.cli.citations import cite_and_print

    try:
        from mace.calculators import mace_mp
    except:
        print(
            "*** Mace is not installed. ***\nPlease install with the following command:\n\n  pip install 'clims[mace]'\n"
        )
        return

    macemp = mace_mp(default_dtype="float64")  # return ASE calculator
    atoms = read(filein, format=format)

    atoms.calc = macemp

    atoms_relax = atoms
    if fix_symmetry:
        atoms.set_constraint(FixSymmetry(atoms))
        atoms_relax = atoms
    if relax_unit_cell and atoms.pbc.all():
        atoms_relax = UnitCellFilter(atoms)

    relax = BFGS(atoms_relax, trajectory=trajectory)

    for ii, _ in enumerate(relax.irun(fmax=fmax)):
        atoms.write(f"geometry.in.next_step.{ii:03}", "aims", scaled=atoms.pbc.all())

    cite_and_print(["ase", "mace-mp-0"])



import click
import numpy as np
from ase.io import read
from clims.phonopy_interface import initialize_phonopy, post_process_phonopy
from clims.cli.citations import cite_and_print

import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)


@click.group()
def clims_phonopy():
    """
    Simple wrapper around phonopy for FHI-aims.
    """


@clims_phonopy.command()
@click.option("--displacement", default=0.01, help="Displacement of atoms (AA).")
@click.option("--supercell", help="Supercell matrix (9 elements)", nargs=9)
@click.option("--filein", help="Input geometry file", default="geometry.in")
def initialize(displacement, supercell, filein):
    """
    Generate geometries with displaced atoms needed
    for the finite difference calculation of phonons DOS.
    """
    supercell_matrix = np.array([int(x) for x in supercell]).reshape((3, 3))
    structure = read(filein)
    initialize_phonopy(structure, supercell_matrix=supercell_matrix, displacement=displacement)
    cite_and_print(["phonopy"])


@clims_phonopy.command()
@click.option(
    "--mesh", default=[20, 20, 20], help="q-point mesh for phonopy", nargs=3, type=int
)
@click.option(
    "--outfile",
    default="aims.out",
    help="FHI-aims output file. All files should have the same name in the subfolders",
)
@click.option("--show", is_flag=True, help="Show the DOS.")
@click.option("--born", is_flag=True, help="Use BORN file for non-analytic correction.")
def postprocess(mesh, outfile, show, born):
    """
    Post-processing the calculated forces to obtain the phonon DOS.
    """
    post_process_phonopy(mesh=mesh, outfile=outfile, show=show, born=born)
    cite_and_print(["ase", "phonopy"])



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import sys
import click
from ase.io import read
from clims.read_dos_data import read_species_dos, read_total_dos, get_dos_groups
from clims.read_band_data import read_bands
from clims.read_control import read_control
from clims.plot_mulliken import plot_mulliken_bands, get_groups_for_projections
from clims.cli.citations import cite_and_print

# from time import time

l_numbers = {"s": 0, "p": 1, "d": 2, "f": 3, "g": 4, "h": 5}
l_marker = ["s", "o", "d", "x", "+", "p"]


@click.command()
@click.option(
    "--emin",
    type=float,
    default=-20.1,
    help="Minimum energy value on the y axis of the plot(s).",
)
@click.option(
    "--emax",
    type=float,
    default=5.1,
    help="Maximum energy value on the y axis of the plot(s).",
)
@click.option(
    "--legend_offset",
    type=(float, float),
    default=None,
    nargs=2,
    help="A x, y offset on the canvas that allows one to shift the legend horizontally.",
)
@click.option("--no_legend", is_flag=True, help="No legend will be printed into the plot.")
@click.option(
    "--show_l_components",
    is_flag=True,
    help="L components of the species_dos will be plotted if available.",
)
@click.option(
    "--species",
    type=str,
    help="Comma-separated list of species groups to plot for the Mulliken projection. The L-projection can be enabled by :l, e.g. --species 'Ga P, As :s'",
)
@click.option(
    "--atoms",
    type=str,
    help="Comma-separated list of atom groups to plot for the Mulliken projection. The L-projection can be enabled by :l, e.g. --atoms '1 2:s'",
)
@click.option(
    "--colors",
    type=str,
    help="Comma-separated list of colors for each defined group. Number of colors must match the number of groups.",
)
@click.option(
    "--bands_plot/--no-bands_plot", default=True, help="Plot/Do not plot Band structure."
)
@click.option(
    "--bands_only", is_flag=True, help="Plot band structure without Mulliken projection."
)
@click.option("--dos_plot/--no-dos_plot", default=True, help="Plot/Do not plot DOS.")
@click.option("--save_only", is_flag=True, help="Only save file.")
@click.option(
    "--fontsize", type=int, default=14, help="Define font size for labels and tick marks."
)
@click.option(
    "--linewidth", type=int, default=1, help="Define line width for band and dos lines."
)
@click.option(
    "--maxdos_output", type=float, default=-1, help="The maximum value of the DOS axis."
)
@click.option(
    "--scale_mulliken_marker",
    type=float,
    default=1,
    help="Scales the marker size in the Mulliken-projected band plot.",
)
def aimsplot(
    emin,
    emax,
    legend_offset,
    no_legend,
    show_l_components,
    species,
    atoms,
    colors,
    bands_plot,
    bands_only,
    dos_plot,
    save_only,
    fontsize,
    linewidth,
    maxdos_output,
    scale_mulliken_marker,
):
    """
    Script to plot band structure and DOS calculated with FHI-aims.
    Requires the control.in/geometry.in files as well as the output
    of the calculation to be in the directory from which the script is called.

    This plot will be created in the same directory as a file aimsplot.png if all goes well.
    Specifically, the script allows to create a plot of

    - Energy band structures                  ("output band ...")

    - Density of states                       ("output dos ...")

    - Species-projected densities of states   ("output species_proj_dos ...")
    as well as, optionally, a decomposition of the DOS into angular momentum components

    - and, alternatively, the tetrahedron-integrated versions of DOS and species-projected DOS

    This script can be called simply as "clims-aimsplot", in which case a default range will be used for
    for the energy range covered on the y axis.

    There are several options that allow one to customize the energy range, type of output,
    legend placement etc.
    There are several options that allow one to customize the energy range, type of output,
    legend placement etc.

    clims-aimsplot --help

    provides an overview of the available options.

    For example,

    clims-aimsplot --emin -20. --emax 10. --legend_offset 1.0 0.2

    will customize both the y axis range shown, as well as the placement of the legend in
    the graph.

    To achieve labelling of the special points along the band structure plot,
    add two arguments to the "output band"
    command in the control.in, using the following syntax:

    output band <start> <end> <npoints> <starting_point_name> <ending_point_name>

    Example: To plot a band with 20 points from Gamma to half way along one of the
           reciprocal lattice vectors, write (in control.in)

    output band 0.0 0.0 0.0 0.5 0.0 0.0 20 Gamma <End_point_name>

    For species-projected DOS and Mulliken-projected band structure plots, it is also
    possible to define individual groups of either species or atoms and optionally assign
    individual colors to the groups. The syntax for the species works as follows:

    clims-aimsplot --species "Pb I, H N C" --colors "blue, green"

    This will create two groups of projections: One for Pb and I, and another for H, N, C. The
    order of the groups also defines the plot order.
    """

    ###########
    # OPTIONS #
    ###########
    # The DPI used for printing out images
    print_resolution = 250
    # Turn on spline interpolation for band structures NOT VERY WELL TESTED!
    # should_spline = False
    # Whether to output the x-axis (e.g. the e=0 line) or not
    output_x_axis = True
    # If spline interpolation turned on, the sampling factor (1 is the original grid)
    # spline_factor = 10

    e_shift = 0

    matplotlib.rcParams["lines.linewidth"] = linewidth
    matplotlib.rcParams["savefig.dpi"] = print_resolution
    matplotlib.rcParams["font.size"] = fontsize

    ########################

    print("Plotting bands for FHI-aims!")
    print("============================\n")
    print("Reading lattice vectors from geometry.in ...")

    structure = read("geometry.in")
    n_atoms = len(structure)
    latvec = structure.get_cell()
    print("Lattice vectors:")
    print(latvec[:], "\n")

    rlatvec = 2.0 * np.pi * latvec.reciprocal()
    print("Reciprocal lattice vectors:")
    print(rlatvec, "\n")

    ########################

    print("Reading information from control.in ...")

    PLOT_BANDS = False
    PLOT_GW = False
    PLOT_BANDS_MULLIKEN = False
    PLOT_DOS = False
    PLOT_DOS_SPECIES = False
    PLOT_DOS_ATOM = False
    PLOT_SOC = False  # This is needed because there will only be one "spin" channel output,
    # but collinear spin may (or may not) be turned on, so the "spin
    # collinear" setting needs to be overridden
    PLOT_DOS_REVERSED = False

    max_spin_channel = 1
    band_segments = []
    band_totlength = 0.0  # total length of all band segments

    try:
        control = read_control()
    except:
        print("Something went wrong while parsing control.in")
        print("The following may missing: control.in or output flags")
        raise

    if atoms and species:
        print("\nPlease specify either '--species' or '--atoms'. Stopping.\n")
        return

    l_index = 0
    if not species and not atoms:
        species = [[s] for s in control["species"].keys()]
        atoms = None
    elif atoms:
        species = None
        if len(atoms.split(":")) == 2:
            atoms, l = atoms.split(":")
            l = l.strip()
            print(f"Plotting atom-projected Output on '{l}'-orbital")
            try:
                l_index = l_numbers[l] + 1
            except KeyError:
                print(
                    f"\nl quantum number '{l}' not in the list {list(l_numbers.keys())}. Stopping.\n"
                )
                return
        elif len(atoms.split(":")) > 2:
            print("Error: Only one l-projection for all groups allowed. Stopping.")
            return
        atoms = atoms.split(",")
        if len(atoms) > 1:
            atoms_tmp = []
            for a in atoms:
                atoms_tmp += [[int(x) for x in a.strip().split()]]
            atoms = atoms_tmp
        else:
            atoms = np.array([atoms[0].split()], dtype=int)
        if max(max(atoms)) > len(structure):
            print("One of your atom number is too big. Stopping.")
            return
    else:
        if len(species.split(":")) == 2:
            species, l = species.split(":")
            l = l.strip()
            print(f"Plotting species-projected Output on '{l}'-orbital")
            try:
                l_index = l_numbers[l] + 1
            except KeyError:
                print(f"l quantum number '{l}' not in the list {list(l_numbers.keys())}")
                return
        elif len(species.split(":")) > 2:
            print("Error: Only one l-projection for all groups allowed. Stopping.")
            return
        species = species.split(",")
        if len(species) > 1:
            species = [s.split() for s in species]
        else:
            species = [species[0].split()]
        atoms = None

    output = control["output"]

    if "spin" in control and control["spin"] == "collinear":
        max_spin_channel = 2

    if (
        "calculate_perturbative_soc" in control
        or "include_spin_orbit" in control
        or "include_spin_orbit_sc" in control
    ):
        PLOT_SOC = True
        max_spin_channel = 1

    if "qpe_calc" in control and control["qpe_calc"] == "gw_expt":
        PLOT_GW = True

    if "bands" in output and bands_plot:
        PLOT_BANDS = True
        if "is_band_mulliken" in output and not bands_only:
            PLOT_BANDS_MULLIKEN = True

        for band in output["bands"]:
            start = np.asarray(band[:3], dtype=float)
            end = np.asarray(band[3:6], dtype=float)
            length = np.linalg.norm(np.dot(end, rlatvec) - np.dot(start, rlatvec))
            band_totlength += length
            npoint = int(band[6])
            startname, endname = "", ""
            if len(band) > 7:
                startname = band[7]
            if len(band) > 8:
                endname = band[8]
            band_segments += [(start, end, length, npoint, startname, endname)]

    if ("dos" in output or "dos_tetrahedron" in output) and dos_plot:
        PLOT_DOS = True

    if (
        ("species_proj_dos" in output or "species_proj_dos_tetrahedron" in output)
        and atoms is None
        and dos_plot
    ):
        PLOT_DOS_SPECIES = True

    if "atom_proj_dos" in output or "atom_proj_dos_tetrahedron" in output:
        PLOT_DOS_ATOM = True

    if PLOT_BANDS_MULLIKEN or PLOT_DOS_SPECIES:
        groups, group_colors, masks = get_groups_for_projections(
            PLOT_SOC, species, atoms, structure, colors
        )

    if PLOT_BANDS and (PLOT_DOS or PLOT_DOS_SPECIES):
        fig, (ax_bands, ax_dos) = plt.subplots(
            nrows=1, ncols=2, layout="tight", sharey=True, width_ratios=[3, 1]
        )
        # ax_bands = plt.axes([0.1, 0.1, 0.6, 0.8])
        # ax_dos = plt.axes([0.72, 0.1, 0.2, 0.8], sharey=ax_bands)
        ax_dos.set_title("DOS")
        plt.setp(ax_dos.get_yticklabels(), visible=False)
        ax_bands.set_ylabel("Energy (eV)")
        PLOT_DOS_REVERSED = True
        plt.setp(ax_bands.spines.values(), lw=2)
        plt.setp(ax_dos.spines.values(), lw=2)
    elif PLOT_BANDS:
        fig, ax_bands = plt.subplots(nrows=1, ncols=1, layout="tight")
        ax_bands.set_ylabel("Energy (eV)")
        plt.setp(ax_bands.spines.values(), lw=2)
    elif PLOT_DOS or PLOT_DOS_SPECIES:
        fig, ax_dos = plt.subplots(nrows=1, ncols=1, layout="tight")
        ax_dos.set_ylabel("DOS")
        plt.setp(ax_dos.spines.values(), lw=2)

    #######################

    if PLOT_BANDS:
        print("Plotting %i band segments..." % len(band_segments))

        if output_x_axis:
            ax_bands.axhline(0, color=(0.0, 0.0, 0.0), linestyle=":")

        band_data, labels, e_shift = read_bands(
            band_segments, band_totlength, max_spin_channel, PLOT_GW
        )
        # start = time()
        for iband, segment in enumerate(band_data.values()):
            for spin in range(max_spin_channel):
                sb = segment[spin]
                ax_bands.plot(sb["xvals"], sb["band_energies"], color=sb["color"])
                if PLOT_BANDS_MULLIKEN:
                    filename = "bandmlk%i%03i.out" % (spin + 1, iband + 1)
                    scatters = plot_mulliken_bands(
                        filename,
                        ax_bands,
                        n_atoms,
                        l_index,
                        PLOT_SOC,
                        sb,
                        masks,
                        groups,
                        group_colors,
                        scale_mulliken_marker,
                    )
        # end = time()
        # print("Timing", end - start)

        if PLOT_BANDS_MULLIKEN and not no_legend:
            lg = ax_bands.legend(
                scatters,
                groups,
                scatterpoints=1,
                fontsize=fontsize,
                loc="lower right",
                bbox_to_anchor=legend_offset,
            )
            for lh in lg.legend_handles:
                lh.set_alpha(1)
                lh.set_sizes([100])

        tickx, tickl = [], []
        for xpos, l in labels:
            ax_bands.axvline(xpos, color="k", linestyle=":")
            tickx += [xpos]
            if l == "Gamma" or l == "G":
                l = "$\\Gamma$"
            tickl += [l]
            print("| %8.3f %s" % (xpos, repr(l)))

        ax_bands.set_xlim(labels[0][0], labels[-1][0])
        ax_bands.set_xticks(tickx)
        ax_bands.set_xticklabels(tickl)
        ax_bands.set_ylim(emin, emax)

    #######################
    maxdos = 0.0
    energy_lim = [100000, -100000]

    if PLOT_DOS_SPECIES:
        dos_species = read_species_dos(list(control["species"].keys()), max_spin_channel)
        dos_groups = get_dos_groups(groups, dos_species, max_spin_channel)
        maxdos = dos_species["maxdos"]
        energy_lim = dos_species["energy_lim"]

    if PLOT_DOS:
        energy, dos, maxdos, energy_lim = read_total_dos()

    spinsgns = [1.0]
    spinlabels = [""]
    if max_spin_channel == 2:
        spinsgns = [1.0, -1.0]
        spinlabels = ["up", "down"]

    if PLOT_DOS_REVERSED:
        ax_dos.axhline(0, color=(1.0, 0.0, 0.0), linestyle=":")
        ax_dos.axvline(0, color=(0.5, 0.5, 0.5))

        if PLOT_DOS:
            for ispin in range(max_spin_channel):
                ax_dos.plot(dos[:, ispin] * spinsgns[ispin], energy + e_shift, color="k")

        if PLOT_DOS_SPECIES:
            for c, s in zip(group_colors, groups):
                for ispin, (spinsgn, spinlabel) in enumerate(zip(spinsgns, spinlabels)):
                    s_energy = dos_groups["x"]
                    species_dos = dos_groups[s][ispin][:, 0]
                    ax_dos.plot(
                        species_dos * spinsgn,
                        s_energy + e_shift,
                        linestyle="-",
                        label=f"{s} {spinlabel}",
                        color=c,
                    )
                    if show_l_components:
                        l_dos = dos_groups[s][ispin][:, 1:]
                        for l in range(l_dos.shape[1]):
                            label = f"{s} (l={l}) {spinlabel}"
                            ax_dos.plot(
                                l_dos[:, l] * spinsgn,
                                s_energy + e_shift,
                                linestyle="--",
                                marker=l_marker[l],
                                markevery=len(s_energy) // 400,
                                label=label,
                                color=c,
                            )

        if maxdos_output > 0:
            # If the user has specified a maximum DOS value, use it
            ax_dos.set_xlim(np.array([min(spinsgns[-1], -0.05), 1.00]) * maxdos_output)
        else:
            # Otherwise use the maximum DOS value read in
            ax_dos.set_xlim(np.array([min(spinsgns[-1], 0.0) - 0.05, 1.05]) * maxdos)

    else:  # not PLOT_DOS_REVERSED
        if PLOT_DOS or PLOT_DOS_SPECIES:
            ax_dos.axvline(0, color="k", ls="--")
            ax_dos.axhline(0, color=(0.5, 0.5, 0.5))
            ax_dos.set_xlabel(r"$\varepsilon - \mu$ (eV)")

        if PLOT_DOS_SPECIES:
            for c, s in zip(group_colors, groups):
                for ispin, (spinsgn, spinlabel) in enumerate(zip(spinsgns, spinlabels)):
                    s_energy = dos_groups["x"]
                    species_dos = dos_groups[s][ispin][:, 0]
                    l_dos = dos_groups[s][ispin][:, 1:]
                    ax_dos.plot(
                        s_energy + e_shift,
                        species_dos * spinsgn,
                        linestyle="-",
                        label=f"{s} {spinlabel}",
                        color=c,
                    )
                    if show_l_components:
                        for l in range(l_dos.shape[1]):
                            label = f"{s} (l={l}) {spinlabel}"
                            ax_dos.plot(
                                s_energy + e_shift,
                                l_dos[:, l] * spinsgn,
                                linestyle="--",
                                marker=l_marker[l],
                                markevery=len(s_energy) // 400,
                                label=label,
                                color=c,
                            )

        if PLOT_DOS:
            for ispin in range(max_spin_channel):
                ax_dos.plot(
                    energy + e_shift, dos[:, ispin] * spinsgns[ispin], label="Total DOS"
                )

            ax_dos.set_xlim(energy[0], energy[-1])
            ax_dos.set_xlim(emin, emax)
            if maxdos_output > 0:
                # If the user has specified a maximum DOS value, use that instead
                ax_dos.set_ylim(np.array([min(spinsgns[-1], -0.05), 1.00]) * maxdos_output)
            else:
                # Otherwise use the maximum DOS value read in
                ax_dos.set_ylim(np.array([min(spinsgns[-1], 0.0) - 0.05, 1.05]) * maxdos)

    if PLOT_DOS_SPECIES and not PLOT_BANDS_MULLIKEN:
        if not no_legend:
            ax_dos.legend(bbox_to_anchor=legend_offset, loc="upper left")

    #######################

    print(
        f"\nThe resolution for saving figures is set to {matplotlib.rcParams['savefig.dpi']} dpi."
    )

    cite_and_print(["aims"])

    def on_q_exit(event):
        if event.key == "q":
            sys.exit(0)

    plt.connect("key_press_event", on_q_exit)
    if save_only:
        plt.savefig("aimsplot.png")
    else:
        plt.show()



from ase.io import read, write
import spglib as spg
import numpy as np
from ase.atoms import Atoms
from ase.cell import Cell


class Structure:
    c_info = []

    def __init__(self, c, sym_thresh=1e-5):

        self.add("Number of atoms", len(c))
        self.add("Chemical formula", c.get_chemical_formula())

        if all(c.pbc):
            lattice = c.get_cell()[:]
            positions = c.get_scaled_positions()
            numbers = c.get_atomic_numbers()
            magmoms = c.get_initial_magnetic_moments()
            cell = (lattice, positions, numbers, magmoms)
            dataset = spg.get_symmetry_dataset(cell, symprec=sym_thresh)
            # Get primitive unit cell
            plattice, pscaled_positions, pnumbers = spg.find_primitive(
                cell, symprec=sym_thresh
            )
            self.primitive = Atoms(
                cell=plattice, pbc=True, scaled_positions=pscaled_positions, numbers=pnumbers
            )
            # Refine cell
            rlattice, rscaled_positions, rnumbers = spg.refine_cell(cell, symprec=sym_thresh)
            self.refined = Atoms(
                cell=rlattice, pbc=True, scaled_positions=rscaled_positions, numbers=rnumbers
            )
            self.add("Unit cell parameters (original)", np.around(c.cell.cellpar(), 10))
            bravais = Cell(lattice).get_bravais_lattice(eps=sym_thresh)
            self.add(
                "Bravais lattice (symmetrized)", "{} {}".format(bravais.longname, bravais)
            )
            self.add("Spacegroup number", dataset.number)
            self.add("Hall symbol", dataset.hall)
            wy = dataset.wyckoffs
            unique_atoms, count = np.unique(dataset.equivalent_atoms, return_counts=True)
            e_atoms = {}
            for ua in unique_atoms:
                e_atoms[ua] = [
                    str(i + 1) for i, ea in enumerate(dataset.equivalent_atoms) if ea == ua
                ]
            if dataset.number > 1:
                self.add(
                    "Occupied Wyckoff positions",
                    f"\n{' '*34}".join(
                        f"{c}*{wy[w]} ({', '.join(e_atoms[w])})"
                        for (w, c) in zip(unique_atoms, count)
                    ),
                )
            self.add("Is primitive cell?", len(c) == len(pnumbers))
            self.add("Symmetry threshold", sym_thresh)

    def __str__(self):
        str_str = "\nStructure Info\n" + "-" * 14 + "\n"
        for info in self.c_info:
            if isinstance(info[1], (list, np.ndarray)):
                fmt_str = "{:32}: " + "{} " * len(info[1]) + "\n"
                str_str += fmt_str.format(info[0], *info[1])
            else:
                fmt_str = "{:32}: {}\n"
                str_str += fmt_str.format(info[0], info[1])
        return str_str

    def add(self, info_str, info_value):
        self.c_info.append([info_str, info_value])

    def write_primitive(self):
        write("geometry-primitive.in", self.primitive, format="aims", scaled=True)

    def write_refined(self):
        write("geometry-conventional.in", self.refined, format="aims", scaled=True)



import click
from ase.io import read
import elastic
from pathlib import Path
from ase.units import GPa
from clims.cli.citations import cite_and_print
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)


@click.group()
def clims_elastic():
    """
    Simple wrapper around Elastic for FHI-aims to compute the elastic tensor and/or the
    equation of state.
    """


@clims_elastic.command()
@click.option(
    "--action",
    type=click.Choice(["cij", "eos"]),
    default="cij",
    help="Generate deformations for elastic constants (cij) or equation of state (eos).",
)
@click.option(
    "--n_deformations",
    default=5,
    type=int,
    help="Number of generated deformations per axis (default: 5)",
)
@click.option(
    "--volume_change",
    default=0.02,
    type=int,
    help="Maximum relative volume change w.r.t V_0.",
)
@click.option(
    "--size",
    default=2.0,
    type=float,
    show_default=True,
    help="Deformation size for Cij scan (% or deg., default: 2.0)",
)
def initialize(action, n_deformations, volume_change, size):
    """
    Generate deformed structures for the computation of the elastic constants and the
    equation of motion with the package 'Elastic'.
    """

    c = read("geometry.in")
    p = Path(".")
    _, brav, sg, _ = elastic.elastic.get_lattice_type(c)
    click.echo(f"{brav} lattice ({sg}): {c.get_chemical_formula()}")
    if action == "cij":
        systems = elastic.get_elementary_deformations(c, n=n_deformations, d=size)
    elif action == "eos":
        systems = elastic.scan_volumes(
            c, n=n_deformations, lo=1 - volume_change, hi=1 + volume_change
        )
    systems.insert(0, c)
    click.echo(f"Writing {len(systems)} deformation files.")
    for n, s in enumerate(systems):
        d = p / f"{action}_{n:03}"
        d.mkdir()
        s.write(d / "geometry.in", scaled=False)

    cite_and_print(["elastic"])


@clims_elastic.command()
@click.option(
    "--action",
    type=click.Choice(["cij", "eos"]),
    default="cij",
    help="Generate deformations for elastic constants (cij) or equation of state (eos).",
)
@click.option("--pwd", type=click.Path(exists=True), default=".")
def postprocess(action, pwd):
    """
    Read the FHI-aims output files from the subdirectories and compute the elastic tensor
    (cij) of the equation of state (eos).
    """

    def reader(f):
        print(f"Reading file {f}")
        return read(f)

    p = Path(pwd)
    outputs = sorted(
        [
            (d / "aims.out").as_posix()
            for d in p.iterdir()
            if d.is_dir() and action in d.as_posix()
        ]
    )
    outputs = [reader(o) for o in outputs]

    if action == "cij":
        cij = elastic.get_elastic_tensor(outputs[0], systems=outputs[1:])
        msv = cij[1][3].max()
        eps = 1e-4
        click.echo("\nCij solution\n" + 30 * "-")
        click.echo(
            f" Solution rank: {cij[1][2]:2d}{' (undetermined)' if cij[1][2] < len(cij[0]) else ''}"
        )
        if cij[1][2] == len(cij[0]):
            click.echo(f" Square of residuals: {cij[1][1]:7.2g}")
        click.echo(" Relative singular values:")
        for sv in cij[1][3] / msv:
            click.echo(f"{sv:7.4f}{'* ' if (sv) < eps else '  '}", nl=False)
        click.echo("\n\nElastic tensor (GPa):")
        for dsc in elastic.elastic.get_cij_order(outputs[0]):
            click.echo(f"{dsc:>7}  ".format(dsc), nl=False)
        click.echo("\n" + 30 * "-")
        for c, sv in zip(cij[0], cij[1][3] / msv):
            click.echo(f"{c / GPa:7.2f}{'* ' if sv < eps else '  '}", nl=False)
        click.echo()
    elif action == "eos":
        eos = elastic.get_BM_EOS(outputs[0], systems=outputs[1:])
        eos[1] /= GPa
        click.echo("# %7s (A^3)  %7s (GPa)   %7s" % ("V0", "B0", "B0'"))
        click.echo("    %7.2f        %7.2f        %7.2f" % tuple(eos))

    cite_and_print(["ase", "elastic"])



def compare_files(file1, file2, comment="#"):
    with open(file1) as f1:
        lines1 = f1.readlines()
    with open(file2) as f2:
        lines2 = f2.readlines()
    assert len(lines1) == len(lines2)
    for l1, l2 in zip(lines1, lines2):
        print(l1, l2)
        if l1.startswith(comment) and l2.startswith(comment):
            pass
        else:
            print(l1, l2)
            assert l1.strip() == l2.strip()


def compare_geo_files(file1, file2, comment="#"):
    with open(file1) as f1:
        lines1 = f1.readlines()
    with open(file2) as f2:
        lines2 = f2.readlines()
    assert len(lines1) == len(lines2)
    for l1, l2 in zip(lines1, lines2):
        if l1.startswith(comment) and l2.startswith(comment):
            pass
        else:
            s_l1 = l1.split()
            s_l2 = l2.split()
            print(l1, l2)
            assert len(s_l1) == len(s_l2)
            if len(s_l1) > 0 and len(s_l2) > 0:
                assert s_l1[0] == s_l2[0]
                assert float(s_l1[1]) == float(s_l1[1])
                assert float(s_l1[2]) == float(s_l1[2])
                assert float(s_l1[3]) == float(s_l1[3])
                if len(s_l1) > 4:
                    assert s_l1[4] == s_l1[4]


def compare_data_files(file1, file2, comment="#"):
    with open(file1) as f1:
        lines1 = f1.readlines()
    with open(file2) as f2:
        lines2 = f2.readlines()
    for l1, l2 in zip(lines1, lines2):
        print(l1, l2)
        if l1.startswith(comment) and l2.startswith(comment):
            pass
        else:
            s_l1 = l1.split()
            s_l2 = l2.split()
            for i1, i2 in zip(s_l1, s_l2):
                assert round(float(i1), 9) == round(float(i2), 9)



import itertools
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection, Line3DCollection
from scipy.spatial import Voronoi


def draw_bound_box(ax, Z):
    # Work around to do axis 'equal'
    max_range = np.amax(np.amax(Z, axis=0) - np.amin(Z, axis=0)) * 1.1
    print(np.amax(Z, axis=0) - np.amin(Z, axis=0))
    center = (np.amax(Z, axis=0) + np.amin(Z, axis=0)) * 0.5
    # print(center)
    # print(max_range)
    box = max_range * np.array(list(itertools.product([0, 1], repeat=3)))
    box -= max_range * 0.5 + center
    # Comment or uncomment following both lines to test the fake bounding box:
    for xb, yb, zb in zip(*box.T):
        ax.plot([xb], [yb], [zb], "w")


def draw_unit_cell(ax, cell):
    uc_corners = np.array(list(itertools.product([0, 1], repeat=3)))
    Z = np.dot(uc_corners, cell)
    verts = [
        [Z[0], Z[1], Z[3], Z[2]],
        [Z[0], Z[1], Z[5], Z[4]],
        [Z[0], Z[2], Z[6], Z[4]],
        [Z[2], Z[3], Z[7], Z[6]],
        [Z[1], Z[3], Z[7], Z[5]],
        [Z[4], Z[5], Z[7], Z[6]],
    ]

    ax.add_collection3d(
        Poly3DCollection(
            verts, facecolors="red", linewidths=1, edgecolors="r", alpha=0.25
        )
    )


def calculate_ws_vertices_from_cell(cell, points):
    """Calculates the Brillouin Zone vertices and path of high-symmetry points.

    Parameters:

    cell: numpy array
        Three lattice vectors arranged as [[lat_v1],[lat_v2],[lat_v3]]
    """

    vor = Voronoi(points)
    bz_vertices = []
    for r in vor.ridge_dict:
        # This is not the general solution. It might happen that for very skewed cells
        # that next-next-nearest neighbors are building the faces of the BZ.
        if r[0] == 62 or r[1] == 62:
            bz_vertices.append([list(vor.vertices[i]) for i in vor.ridge_dict[r]])

    return bz_vertices


def get_equation_of_plane(vertex_list):
    p = np.array(vertex_list)
    n = np.cross(p[1] - p[0], p[2] - p[0])
    n = n / np.linalg.norm(n)
    d = -np.dot(p[0], n)
    return np.array([*n, d])


def get_face_center(p):
    return np.array(p).sum(axis=0) / 4


def plot_plane_normal(abcd, onpoint, ax):
    n = abcd[:3]
    ax.quiver(*onpoint, *n, length=5, normalize=True)


def get_shortest_vector_to_plane(plane, point):
    k = (np.dot(plane[:3], point) + plane[3]) / np.linalg.norm(plane[:3]) ** 2.0
    c_point = point - k * plane[:3]
    d_vec = -c_point + point
    # ax.quiver(*c_point, *d_vec)
    # ax.quiver(*c_point, *plane[:3])
    return d_vec


def check_if_point_in_polyhedron(point, planes):
    is_in = []
    # print(len(planes))
    for plane in planes:
        d_vec = get_shortest_vector_to_plane(plane, point)
        if np.linalg.norm(d_vec) < 1e-8:
            is_in.append(True)
        else:
            unit_d_vec = d_vec / np.linalg.norm(d_vec)
            # print(np.linalg.norm(-plane[:3] - unit_d_vec),d_vec)
            if np.linalg.norm(-plane[:3] - unit_d_vec) < 1e-8:
                is_in.append(True)
            else:
                # print('No',plane[:3] - unit_d_vec)
                assert np.linalg.norm(plane[:3] - unit_d_vec) < 1e-8
                is_in.append(False)
    return all(is_in)


def move_atom_in_ws(pos, planes, points):
    for p in points:
        new_pos = pos + p
        # print("Suggest new position:",new_pos, pos, p)
        is_in = check_if_point_in_polyhedron(new_pos, planes)
        if is_in:
            # print('New position found')
            return new_pos
    raise


def get_ws_planes(ws_vertices):
    planes = []
    for vertex in ws_vertices:
        abcd = get_equation_of_plane(vertex)
        onpoint = get_face_center(vertex)
        # print(np.linalg.norm(abcd[:3] - onpoint/np.linalg.norm(onpoint)))
        if not np.linalg.norm(abcd[:3] - onpoint / np.linalg.norm(onpoint)) < 1e-8:
            abcd *= -1.0
        # print(np.linalg.norm(abcd[:3] - onpoint/np.linalg.norm(onpoint)))
        assert np.linalg.norm(abcd[:3] - onpoint / np.linalg.norm(onpoint)) < 1e-8
        planes.append(abcd)
        # plot_plane_normal(abcd,onpoint,ax)
    return planes


def create_ws_cell(structure):
    cell = structure.get_cell()[:]
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    corners = np.array(list(itertools.product([-2.0, -1.0, 0.0, 1.0, 2.0], repeat=3)))
    points = np.dot(corners, cell)

    # draw_unit_cell(ax,cell)
    ws_vertices = calculate_ws_vertices_from_cell(cell, points)

    # ax.add_collection3d(Poly3DCollection(
    #    ws_vertices, facecolors='cyan', linewidths=1, edgecolors='r', alpha=.25)
    # )

    planes = get_ws_planes(ws_vertices)

    for i, pos in enumerate(structure.get_positions()):
        # print('--Atom', i, pos)
        is_in = check_if_point_in_polyhedron(pos, planes)
        # print(is_in)
        if not is_in:
            new_pos = move_atom_in_ws(pos, planes, points)
            structure[i].position = new_pos

    # ax.scatter3D(*crystal.get_positions().T)
    # draw_bound_box(ax,points)

    # plt.show()



from click.testing import CliRunner
from clims.cli.prepare_run import prepare_run
from ase.build import bulk
from ase.io import write
from test.file_comparison import compare_files
from pathlib import Path
from clims.cli.configure import get_config_value, clims_configure

GaAs = bulk("GaAs", crystalstructure="zincblende", a=5.65315)


def test_prepare_run_default():
    runner = CliRunner()
    try:
        get_config_value("species_path")
    except:
        species_path = Path.cwd() / "test" / "species_defaults"
        runner.invoke(clims_configure, ["--species-path", species_path.as_posix()])

    with runner.isolated_filesystem():
        with open("control.in_ref", "w") as f:
            f.write(control_in)
        write("geometry.in", GaAs, scaled=True, format="aims")
        result = runner.invoke(prepare_run, ["--bands", "--relax"])
        assert result.exit_code == 0
        compare_files("control.in", "control.in_ref")


control_in = """\
#===============================================================================
# Created using the Atomic Simulation Environment (ASE)

# Fri Apr 23 17:56:41 2021

#===============================================================================
xc                                 pbe
relativistic                       atomic_zora scalar
k_grid                             12 12 12
relax_geometry                     trm 5e-3
relax_unit_cell                    full
output                             band   0.00000  0.00000  0.00000   0.50000  0.00000  0.50000   39 G  X
output                             band   0.50000  0.00000  0.50000   0.50000  0.25000  0.75000   19 X  W
output                             band   0.50000  0.25000  0.75000   0.37500  0.37500  0.75000   14 W  K
output                             band   0.37500  0.37500  0.75000   0.00000  0.00000  0.00000   41 K  G
output                             band   0.00000  0.00000  0.00000   0.50000  0.50000  0.50000   34 G  L
output                             band   0.50000  0.50000  0.50000   0.62500  0.25000  0.62500   24 L  U
output                             band   0.62500  0.25000  0.62500   0.50000  0.25000  0.75000   14 U  W
output                             band   0.50000  0.25000  0.75000   0.50000  0.50000  0.50000   28 W  L
output                             band   0.50000  0.50000  0.50000   0.37500  0.37500  0.75000   24 L  K
output                             band   0.62500  0.25000  0.62500   0.50000  0.00000  0.50000   14 U  X
#===============================================================================

################################################################################
#
#  FHI-aims code project
#  VB, Fritz-Haber Institut, 2009
#
#  Suggested "light" defaults for Ga atom (to be pasted into control.in file)
#  Be sure to double-check any results obtained with these settings for post-processing,
#  e.g., with the "tight" defaults and larger basis sets.
#
#  2020/09/15 Added f function to light settings to improve bulk properties of Ga-Br.
#
################################################################################
  species        Ga
#     global species definitions
    nucleus             31
    mass                69.723
#
    l_hartree           4
#
    cut_pot             3.5          1.5  1.0
    basis_dep_cutoff    1e-4
#
    radial_base         54 5.0
    radial_multiplier   1
    angular_grids       specified
      division   0.5103   50
      division   0.8880  110
      division   1.2009  194
      division   1.5000  302
#      division   1.7093  434
#      division   1.8791  590
#      division   1.9525  770
#      division   2.3801 1202
#      outer_grid  1202
      outer_grid  302
################################################################################
#
#  Definition of "minimal" basis
#
################################################################################
#     valence basis states
    valence      4  s   2.
    valence      4  p   1.
    valence      3  d  10.
#     ion occupancy
    ion_occ      4  s   1.
    ion_occ      3  p   6.
    ion_occ      3  d  10.
################################################################################
#
#  Suggested additional basis functions. For production calculations, 
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  Constructed for dimers: 1.85 A, 2.10 A, 2.45 A, 3.00 A, 4.00 A
#
################################################################################
#  "First tier" - improvements: -222.33 meV to -26.19 meV 
     hydro 2 p 1.2
     hydro 3 d 3.8
     hydro 4 f 6.8
     ionic 4 s auto
#  "Second tier" - improvements: -11.68 meV to -1.61 meV
#     hydro 5 g 10
#     hydro 4 p 3.6
#     hydro 4 f 13.2
#     hydro 6 h 14.4
#     hydro 4 d 5.2
#     hydro 1 s 0.45
#  "Third tier" - improvements: -0.64 meV to -0.15 meV
#     hydro 3 p 3.4
#     hydro 3 s 2.2
#     hydro 5 g 14
#     hydro 4 f 6.2
#     hydro 5 d 7.2
#  "Fourth tier"  -improvements: -0.12 meV and below
#     hydro 3 s 3.8
#     hydro 5 f 27.2
#     hydro 6 h 16
#     hydro 5 g 9.2
#     hydro 4 d 8.6
#     hydro 2 p 3.6  

################################################################################
#
#  FHI-aims code project
#  VB, Fritz-Haber Institut, 2009
#
#  Suggested "light" defaults for As atom (to be pasted into control.in file)
#  Be sure to double-check any results obtained with these settings for post-processing,
#  e.g., with the "tight" defaults and larger basis sets.
#
#  2020/09/15 Added f function to light settings to improve bulk properties of Ga-Br.
#
################################################################################
  species        As
#
    nucleus             33
    mass                74.92160
#
    l_hartree           4
#
    cut_pot             3.5          1.5  1.0
    basis_dep_cutoff    1e-4
#
    radial_base         55 5.0
    radial_multiplier   1
    angular_grids       specified
      division   0.4982   50
      division   0.9113  110
      division   1.1593  194
      division   1.4959  302
#      division   1.6697  434
#      division   1.8319  590
#      division   1.9752  770
#      division   2.0131  974
#      division   2.4015 1202
#      outer_grid  1202
      outer_grid  302
################################################################################
#
#  Definition of "minimal" basis
#
################################################################################
#     valence basis states
    valence      4  s   2.
    valence      4  p   3.
    valence      3  d  10.
#     ion occupancy
    ion_occ      4  s   1.
    ion_occ      4  p   2.
    ion_occ      3  d  10.
################################################################################
#
#  Suggested additional basis functions. For production calculations, 
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  Constructed for dimers: 1.75 A, 2.10 A, 2.50 A, 3.00 A, 4.00 A
#
################################################################################
#  "First tier" - improvements: -385.12 meV to -54.94 meV 
     hydro 3 d 4
     hydro 2 p 1.5
     hydro 4 f 6.8
     ionic 4 s auto
#  "Second tier" - improvements: -22.14 meV to -3.21 meV
#     hydro 5 g 10
#     hydro 6 h 13.6
#     hydro 4 p 4.3
#     hydro 4 f 15.6
#     hydro 4 d 5.4
#     hydro 1 s 0.6
#  "Third tier" - improvements: -0.57 meV to -0.13 meV
#     hydro 5 g 16.4
#     hydro 4 f 7.4
#     hydro 5 d 7.4
#     ionic 4 p auto
#     hydro 3 s 2.6     
#  "Fourth tier" - improvements: -0.30 meV and below
#     hydro 3 p 3
#     hydro 6 h 18.4
#     hydro 5 d 11.2
#     hydro 5 f 15.2
#     hydro 5 g 13.6
#     hydro 5 s 6.2  
"""



import matplotlib.pyplot as plt
import click
import numpy as np

@click.command()
@click.argument("xyfiles", nargs=-1)
def xyplot(xyfiles):
    """
    Simple X-Y plot of data files with first column represent X axis and second and
    following columns represent Y data.
    """ 
    for i, xyfile in enumerate(xyfiles):
        x, *ys = np.loadtxt(xyfile).T
        for j, y in enumerate(ys):
            label = f'File[{i}]: y[{j}]'
            plt.plot(x,y,label=label)
    plt.legend()
    plt.show()



def read_control():
    control = {"species": {}}
    output = {}
    bands = []
    with open("control.in", "r") as f:
        lines = f.readlines()
    # Just read twice: first keywords, second species
    for line in lines:
        words = line.split()
        if len(words) == 0 or words[0].startswith("#"):
            pass
        elif words[0] == "output":
            if words[1] == "band" or words[1] == "band_mulliken":
                bands += [" ".join(words[2:]).split("#")[0].split()]
                if words[1] == "band_mulliken":
                    output["is_band_mulliken"] = True
            else:
                if words[1] == "species_proj_dos":
                    control["n_points_species_dos"] = int(words[4])
                output[words[1]] = " ".join(words[2:]).split("#")[0].split()
        elif words[0] == "species":
            break
        else:
            control[words[0]] = " ".join(words[1:]).split("#")[0]
    if len(bands) > 0:
        output["bands"] = bands
    control["output"] = output

    for line in lines:
        words = line.split()
        if len(words) == 0 or words[0].startswith("#"):
            pass
        elif words[0] == "species":
            control["species"][words[1]] = {}

    return control



import click
from clims.prepare_input import write_control
from pathlib import Path
from clims.cli.configure import get_config_value
from clims.prepare_submission import write_submission_file
from ase.io import read
from clims.cli.citations import cite_and_print


@click.command()
@click.option("--species", default="light", help="Specify species defaults", show_default=True)
@click.option("--relax", is_flag=True, help="Set relaxation flag.")
@click.option("--bands", is_flag=True, help="Attach band path.")
@click.option("--bands_mulliken", is_flag=True, help="Attach Mulliken-decomposed band path.")
@click.option("--dos", is_flag=True, help="Attach DOS output.")
@click.option("--submit-file", is_flag=True, help="Write submission file.")
@click.option("--hse06", is_flag=True, help="Use xc HSE06.")
@click.option("--vdwTS", is_flag=True, help="Use vdW-TS corrections.")
@click.option("--forces", is_flag=True, help="Compute forces.")
@click.option("--stress", is_flag=True, help="Compute (analytical) stress.")
@click.option(
    "--keyword", type=(str, str), multiple=True, help="Add additional keywords to control.in."
)
def prepare_run(
    species,
    relax,
    bands,
    bands_mulliken,
    dos,
    submit_file,
    hse06,
    vdwts,
    forces,
    stress,
    keyword,
):
    """Prepares control.in template file in the CWD, if geometry.in is in CWD

    Provides only a minimum of needed flags to make it run. Following scenarios are coverd:

    1. No control.in file: Creates a new file with minimum flags and species attached\n
    2. control.in file with flags: Attaches the needed species\n
    3. control.in file with species: do nothing\n
    """
    try:
        AIMS_SPECIES = Path(get_config_value("species_path"))
    except:
        return
    species_dir = AIMS_SPECIES / species
    print(species_dir)
    try:
        geo = read("geometry.in", format="aims")
        print("-- Found geometry.in file")
    except:
        print("**ERROR**: No control.in file generated: No geometry.in found!")
        raise
    CWD = Path(".")
    # more kw
    if keyword:
        kw = dict(keyword)
    else:
        kw = {}
    if vdwts:
        kw.update({"vdw_correction_hirshfeld": True})
    if forces:
        kw.update({"compute_forces": True})
    if stress:
        kw.update({"compute_analytical_stress": True})
    write_control(
        geo,
        CWD,
        species_dir,
        relax=relax,
        bands=bands,
        bands_mulliken=bands_mulliken,
        dos=dos,
        hse06=hse06,
        **kw
    )

    if submit_file:
        header = get_config_value("submission_header")
        mpi_command = get_config_value("mpi_command")
        aims_executable = get_config_value("aims_executable")
        output_name = get_config_value("output_name")
        error_name = get_config_value("error_name")
        if all([header, mpi_command, aims_executable, output_name, error_name]):
            write_submission_file(
                header, mpi_command, aims_executable, output_name, error_name
            )
        else:
            print(
                "Some configurations missing. Please check if you have configured the following:"
            )
            print("  submission_header, mpi_command, aims_executable, output_name, error_name")

    cite_and_print(["ase"])



from clims.construct_wigner_seitz_cluster import create_ws_cell
from ase.build.supercells import make_supercell
from ase.io import read, write
import click
import numpy as np


@click.command()
@click.argument("super_matrix", nargs=-1)
@click.option(
    "--filein",
    type=click.Path(exists=True),
    default="geometry.in",
    show_default=True,
    help="File name of input structure",
)
@click.option(
    "--format", type=str, default="aims", help="Format of the in- and output file"
)
@click.option(
    "--fileout",
    default="geometry-ws-cluster.in",
    show_default=True,
    help="File name for cluster structure",
)
@click.option(
    "--translate-basis",
    type=float,
    default=[0.0, 0.0, 0.0],
    nargs=3,
    help="Translates the basis atoms by a vector (in AA). ",
)
def construct(super_matrix, translate_basis, filein, fileout, format):
    """
    Constructs a Cluster of the shape of the Wigner-Seitz Cell based on the supercell
    (of size (det(SUPER_MATRIX))) of a periodic input structure.
    """
    prim = read(filein, format=format)
    prim.positions += np.array(translate_basis)
    if not all(prim.get_pbc()):
        click.echo("\nLoaded structure is not periodic.")
        return

    click.echo(f"\n{'Read structure from:':35}: {filein}")
    click.echo(f"{'Chemical formula':35}: {prim.get_chemical_formula()}")
    if len(super_matrix) == 3:
        P = np.diag([int(i) for i in super_matrix])
    elif len(super_matrix) == 9:
        P = np.array([int(i) for i in super_matrix], dtype=np.double).reshape(3, 3)
    else:
        click.echo("\n***ERROR***: The size of SUPER_MATRIX should be either 3 or 9.")
        return
    supercell = make_supercell(prim, P, wrap=False)
    create_ws_cell(supercell)
    supercell.pbc = False
    write(fileout, supercell, format=format)



citations = {
    "aims": """\
Volker Blum, Ralf Gehrke, Felix Hanke, Paula Havu, Ville Havu,
Xinguo Ren, Karsten Reuter, and Matthias Scheffler
Ab initio molecular simulations with numeric atom-centered orbitals
Computer Physics Communications 180, 2175-2196 (2009)
http://doi.org/10.1016/j.cpc.2009.06.022
""",
    "ase": """\
Larsen, Ask Hjorth, et al.
The Atomic Simulation Environment—A Python library for working with atoms
J. Phys.: Condens. Matter 29, 273002 (2017)
https://doi.org/10.1088/1361-648X/aa680e
""",
    "elastic": """\
Jochym, Pawel T.
Module for calculating elastic tensor of crystals
software, https://github.com/jochym/Elastic
https://doi.org/10.5281/zenodo.593721
""",
    "mace-mp-0": """\
Batatia, Ilyes, et al. 
A foundation model for atomistic materials chemistry 
arXiv preprint arXiv:2401.00096 (2023)
https://doi.org/10.48550/arXiv.2401.00096
""",
    "phonopy": """\
Togo, Atsushi and Chaput, Laurent and Tadano, Terumasa and Tanaka, Isao
Implementation strategies in phonopy and phono3py
J. Phys. Condens. Matter 35, 353001-1-22 (2023)
https://dx.doi.org/10.1088/1361-648X/acd831
""",
    "spglib": """\
Togo, Atsushi  and Tanaka, Isao 
Spglib: a software library for crystal symmetry search
arXiv preprint arXiv:1808.01590 (2018)
https://arxiv.org/abs/1808.01590
""",
}


def cite_and_print(refs):
    print("\nPlease consider to cite the following publications:\n")
    for ref in refs:
        lines = citations[ref].splitlines()
        cite_text = f"{f'[{ref}]':11} " + lines[0] + "\n"
        for l in lines[1:]:
            cite_text += " " * (12) + l + "\n"
        print(cite_text)



import click
from ase.io import read, write
from ase.io.formats import filetype
from clims.cli.citations import cite_and_print


@click.command()
@click.option(
    "--filein",
    type=click.Path(exists=True),
    default="geometry.in",
    show_default=True,
    help="File of input structure",
)
@click.option(
    "--formatin",
    type=str,
    show_default=True,
    help="Format of input file.",
)
@click.option(
    "--fileout",
    type=str,
    default="geometry-converted.in",
    show_default=True,
    help="File name for converted structure",
)
@click.option(
    "--formatout",
    type=str,
    show_default=True,
    help="Format of output file.",
)
@click.option(
    "--fractional", is_flag=True, help="Convert input file to fractional coordinates."
)
@click.option("--cartesian", is_flag=True, help="Convert input file to cartesian coordinates.")
def convert(filein, formatin, fileout, formatout, fractional, cartesian):
    """Convert input file to specified output file."""
    structure = read(filein, formatin)
    scaled = None
    kwargs = {}
    if (fractional is False) and (cartesian is False) and all(structure.pbc):
        scaled = True
    if fractional:
        scaled = True
    if formatout is None:
        if filetype(fileout, read=False) == "aims":
            kwargs["scaled"] = scaled
        if filetype(fileout, read=False) == "vasp":
            kwargs["direct"] = scaled
    write(fileout, structure, format=formatout, **kwargs)

    cite_and_print(["ase"])



import click
import json
from pathlib import Path

config_dir = Path.home() / ".local" / "etc"
config_file = config_dir / "clims_config.json"


@click.command()
@click.option(
    "--species-path",
    type=click.Path(exists=True),
    help="Path to the aims species_defaults folder.",
)
@click.option(
    "--aims-executable",
    type=click.Path(exists=True),
    help="Path to the the aims executable.",
)
@click.option(
    "--submission-header",
    default=None,
    type=click.File(),
    help="File containing a submission header for queing systems.",
)
@click.option(
    "--mpi-command", default=None, type=str, help="MPI run command, e.g. 'mpirun -n 4'."
)
@click.option(
    "--output-name", default=None, type=str, help="Name of the aims output file."
)
@click.option("--error-name", default=None, type=str, help="Name of the error file.")
@click.option(
    "--submit-command",
    default=None,
    type=str,
    help="Command to submit a calculation to queing system.",
)
@click.option("--show", is_flag=True, help="Show the content of the configure file.")
def clims_configure(
    species_path,
    aims_executable,
    mpi_command,
    output_name,
    error_name,
    submission_header,
    submit_command,
    show,
):
    """
    Configure clims standard parameters. This is mandatory for some of the cli tools,
    e.g. `clims-prepare_run` , etc.
    """

    try:
        with open(config_file) as config_json:
            config = json.load(config_json)
    except FileNotFoundError:
        config = {}
        if not config_dir.exists():
            config_dir.mkdir(parents=True)

    if show:
        print("-- Configuration file:")
        for key, value in config.items():
            print(f"{key:20} : {value}")
        return

    # First print the current config file

    if species_path:
        s_path = Path(species_path)
        config["species_path"] = s_path.as_posix()

    if aims_executable:
        a_path = Path(aims_executable)
        config["aims_executable"] = a_path.as_posix()

    if submission_header:
        sh_text = submission_header.read()
        config["submission_header"] = sh_text

    if output_name:
        config["output_name"] = output_name

    if error_name:
        config["error_name"] = error_name

    if mpi_command:
        config["mpi_command"] = mpi_command

    if submit_command:
        config["submit_command"] = submit_command

    with open(config_file, "w") as config_json:
        json.dump(config, config_json)


def get_config_value(key):
    """Return the `value` of `key` of `config_json` file."""

    try:
        with open(config_file) as config_json:
            config = json.load(config_json)
            value = config[key]
            return value
    except FileNotFoundError:
        print("A config value was requested. Please configure clims first!")
        print('Get more information with: "clims-configure --help"')
        raise
    except KeyError:
        print("Default {key} is not set, yet.")
        print(f'Please set with: "clims-configure --{key.replace("_","-")} <...>"')
        raise



import numpy as np
from clims.elements import colors as element_colors
import re
from ase.symbols import symbols2numbers
from ase.data import atomic_numbers
import sys
from matplotlib.colors import is_color_like


def plot_mulliken_bands(
    filename,
    ax_bands,
    n_atoms,
    l_index,
    is_SOC,
    segments,
    masks,
    groups,
    group_colors,
    scale_marker,
):

    markersizeunit = 24
    factor_atoms = 1
    total_atom_ind = 4 + l_index

    if is_SOC:
        factor_atoms = 2
        total_atom_ind += 1

    with open(filename) as f:
        output = f.read()
    print("Parsing " + filename)
    k_points = output.split("k point number:")[1:]
    # print("Number of k_points", len(k_points))
    len_k = len(k_points)

    # Find out actual number of states:
    start_state = int(re.search(r"\n *State *([0-9]*)\n", k_points[0]).group(1)) - 1
    states = re.split(r"\n *State *[0-9]*\n", k_points[0])[1:]
    n_states = len(states)
    mulliken_k = np.zeros((len(groups), n_states * len_k))
    x_vals = np.zeros(n_states * len_k)
    y_vals = np.zeros(n_states * len_k)
    raise_index_warning = False
    for i_k, k in enumerate(k_points):
        start_state = int(re.search(r"\n *State *([0-9]*)\n", k).group(1)) - 1
        states = re.split(r"\n *State *[0-9]*\n", k)[1:]
        end_state = start_state + n_states
        mulliken = np.zeros((n_states, factor_atoms * n_atoms))
        for i_state, state in enumerate(states):
            lines = state.splitlines()
            for i, line in enumerate(lines):
                try:
                    mulliken[i_state, i] = float(line.split()[total_atom_ind])
                except IndexError:
                    mulliken[i_state, i] = 0.0
                    raise_index_warning = True

        for i_s, s in enumerate(groups):
            spec_mlk = mulliken[:, masks[s]].sum(axis=1)
            mulliken_k[i_s, i_k * n_states : (i_k + 1) * n_states] = spec_mlk
            x_vals[i_k * n_states : (i_k + 1) * n_states] = [segments["xvals"][i_k]] * n_states
            y_vals[i_k * n_states : (i_k + 1) * n_states] = segments["band_energies"][i_k][
                start_state:end_state
            ]
    if raise_index_warning:
        print("WARNING: The requested l-projection was not available for at least one atom!")
    scatters = ()

    for i_s, color in enumerate(group_colors):
        sc = ax_bands.scatter(
            x_vals,
            y_vals,
            s=(markersizeunit * np.abs(mulliken_k[i_s])) * scale_marker,
            c=color,
            alpha=1,
            edgecolors="none",
        )
        scatters += (sc,)

    return scatters


def get_groups_for_projections(is_SOC, species, atoms, structure, colors):
    masks = {}
    group_colors = []
    groups = []

    if species is not None:
        for s in species:
            groups += [" ".join(s)]
            for x in s:
                if not x in structure.get_chemical_symbols():
                    print(f"Error: Species '{x}' is not in the structure. Stopping.")
                    sys.exit()
            group_colors += [element_colors[symbols2numbers(s[0])[0] - 1]]
            if is_SOC:
                # There are double the number of states for SOC.
                masks[groups[-1]] = [
                    np.any(np.array(s) == x)
                    for x in structure.get_chemical_symbols()
                    for _ in (0, 1)
                ]
            else:
                masks[groups[-1]] = [
                    np.any(np.array(s) == x) for x in structure.get_chemical_symbols()
                ]
    elif atoms is not None:
        for a in atoms:
            groups += ["Atom No: " + ", ".join([str(x) for x in a])]
            group_colors += [
                element_colors[
                    symbols2numbers(structure.get_chemical_symbols()[a[0] - 1])[0] - 1
                ]
            ]
            if is_SOC:
                masks[groups[-1]] = [
                    (i + 1) in a for i in range(len(structure)) for _ in (0, 1)
                ]
            else:
                masks[groups[-1]] = [(i + 1) in a for i in range(len(structure))]

    if colors:
        colors = [c.strip() for c in colors.split(",")]
        for c in colors:
            if not is_color_like(c):
                print(
                    f"Error: The color '{c}' is not recognized as color by matplotlib. Stopping."
                )
                sys.exit()
        if len(colors) == len(group_colors):
            group_colors = colors
        else:
            print("Error: Number of colors does not equal the number of groups. Stopping.")
            sys.exit()

    return groups, group_colors, masks



from ase.spacegroup import crystal
from click.testing import CliRunner
from clims.cli.unit_cell_info import unit_cell_info
from ase.io import write, read
import numpy as np
from pathlib import Path

Fe2O3 = crystal(
    ("Fe", "O"),
    basis=[
        (0.3333333333333334, 0.6666666666666667, 0.3116866666666667),
        (0.0251333333333334, 0.3584666666666668, 0.4166666666666667),
    ],
    cellpar=[5.0342021271, 5.0342021271, 13.75191, 90.0, 90.0, 120.0],
    spacegroup=167,
    primitive_cell=True,
)

unit_cell_info_output = """Writing conventional unit cell (refined)
Writing primitive unit cell (refined)

Structure Info
--------------
Number of atoms                 : 10
Chemical formula                : Fe4O6
Unit cell parameters (original) : 5.4277538 5.4277538 5.4277538 55.2582374093 55.2582374093 55.2582374093 
Bravais lattice (symmetrized)   : primitive rhombohedral RHL(a=5.42775, alpha=55.2582)
Spacegroup number               : 167
Hall symbol                     : -R 3 2"c
Occupied Wyckoff positions      : 4*c (1, 2, 7, 8)
                                  6*e (3, 4, 5, 6, 9, 10)
Is primitive cell?              : True
Symmetry threshold              : 1e-05


Please consider to cite the following publications:

[ase]       Larsen, Ask Hjorth, et al.
            The Atomic Simulation Environment—A Python library for working with atoms
            J. Phys.: Condens. Matter 29, 273002 (2017)
            https://doi.org/10.1088/1361-648X/aa680e

[spglib]    Togo, Atsushi  and Tanaka, Isao 
            Spglib: a software library for crystal symmetry search
            arXiv preprint arXiv:1808.01590 (2018)
            https://arxiv.org/abs/1808.01590

"""


def test_unit_cell_info():
    runner = CliRunner()
    with runner.isolated_filesystem():
        write("geometry.in", Fe2O3, scaled=True)
        result = runner.invoke(unit_cell_info, ["--conventional", "--primitive"])
        assert result.exit_code == 0
        assert Path("geometry-primitive.in").exists()
        assert Path("geometry-conventional.in").exists()
        conventional = read("geometry-conventional.in")
        primitive = read("geometry-primitive.in")
        assert len(conventional) == 30
        assert len(primitive) == 10
        assert result.output == unit_cell_info_output



from ase.io import read, write
from ase import Atoms
from phonopy.structure.atoms import PhonopyAtoms
from phonopy import Phonopy
from pathlib import Path
import numpy as np


def write_phonopy_disp_yaml(phonon, calculator, supercell_matrix):

    from phonopy.interface.calculator import get_default_physical_units
    from phonopy.interface.phonopy_yaml import PhonopyYaml

    filename = "phonopy_disp.yaml"
    yaml_settings = {
        "force_sets": False,
        "force_constants": False,
        "born_effective_charge": False,
        "dielectric_constant": False,
        "displacements": True,
    }
    adim = [str(x) for x in np.array(supercell_matrix).flatten()]
    dim = " ".join(adim)

    phonopy_confs = {
        "dim": dim,
        "calculator": calculator,
        "create_displacements": ".true.",
    }

    units = get_default_physical_units(calculator)

    phpy_yaml = PhonopyYaml(
        configuration=phonopy_confs, physical_units=units, settings=yaml_settings
    )
    phpy_yaml.set_phonon_info(phonon)
    with open(filename, "w") as w:
        w.write(str(phpy_yaml))


def initialize_phonopy(
    structure,
    structure_file="geometry.in",
    structure_format="aims",
    supercell_matrix=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    primitive_matrix="auto",
    displacement=0.01,
):

    atoms = PhonopyAtoms(
        symbols=structure.get_chemical_symbols(),
        cell=structure.get_cell(),
        scaled_positions=structure.get_scaled_positions(),
    )

    phonon = Phonopy(atoms, supercell_matrix, primitive_matrix=primitive_matrix)
    phonon.generate_displacements(distance=displacement)
    supercells = phonon.get_supercells_with_displacements()

    n_digits = len(str(len(supercells))) + 1
    p = Path(".")
    for i, supercell in enumerate(supercells):
        atoms = Atoms(
            symbols=supercell.get_chemical_symbols(),
            scaled_positions=supercell.get_scaled_positions(),
            cell=supercell.get_cell(),
            pbc=True,
        )
        displacement = p / f"displacement-{str(i).zfill(n_digits)}"
        displacement.mkdir(exist_ok=True)
        print(f"Creating displacement in directory: {displacement.as_posix()}")
        write(displacement / structure_file, atoms, format=structure_format)

    write_phonopy_disp_yaml(phonon, structure_format, supercell_matrix)


def parse_forces(num_atoms, outfile):
    from phonopy.interface.aims import parse_set_of_forces

    p = Path(".")
    output_files = [str(s) for s in p.glob(f"**/{outfile}")]
    print("Found following output files:")
    for output_file in output_files:
        print(output_file)
    forces = parse_set_of_forces(num_atoms, output_files)
    return forces


def post_process_phonopy(mesh=[20, 20, 20], outfile="aims.out", show=False, born=False):
    from phonopy import load
    from phonopy.file_IO import get_born_parameters
    from phonopy.phonon.band_structure import get_band_qpoints_and_path_connections
    from ase.dft.kpoints import resolve_kpt_path_string

    phonon = load(phonopy_yaml="phonopy_disp.yaml", produce_fc=False, is_nac=False)
    primitive = phonon.primitive

    # Parse forces from aims.out files
    n_atoms = phonon.supercell.get_number_of_atoms()
    forces = parse_forces(n_atoms, outfile)
    n_atoms_supercell = len(forces[0])
    assert n_atoms == n_atoms_supercell

    # Do we need non-analytic correction (polar materials)? Yes -> read BORN file
    if born:
        nac_params = get_born_parameters(open("BORN"), primitive, phonon.primitive_symmetry)
        if nac_params is None:
            print("\n***ERROR***: Please correct BORN file. Stopping.\n")
            return
        phonon.set_nac_params(nac_params)

    # Prepare displacement_dataset for phonopy and produce force constants
    first_atoms = []
    for fs, disp in zip(forces, phonon.get_displacements()):
        first_atoms.append({"number": disp[0], "displacement": disp[1:], "forces": fs})
    displacement_dataset = {"natom": n_atoms_supercell, "first_atoms": first_atoms}
    phonon.dataset = displacement_dataset
    phonon.produce_force_constants()

    # Run and write phonon DOS
    phonon.run_mesh(mesh)
    phonon.run_total_dos()
    phonon.write_total_dos()

    # Run and write phonon band structure
    # We are using tools from ASE, which is why we construct atoms object first
    atoms = Atoms(
        symbols=primitive.get_chemical_symbols(),
        cell=primitive.get_cell(),
        positions=primitive.get_positions(),
        pbc=True,
    )
    bp = atoms.cell.bandpath()
    labels, bands = resolve_kpt_path_string(bp.path, bp.special_points)
    labels = [
        l if l != "G" else "$\\Gamma$" for label in labels for l in label
    ]  # need to flatten the array and replace G with Latex Gamma
    print(labels)
    qpoints, connections = get_band_qpoints_and_path_connections(bands, npoints=51)
    phonon.run_band_structure(qpoints, path_connections=connections, labels=labels)
    phonon.write_yaml_band_structure()

    # Show phonon DOS and band structure
    if show:
        phonon.plot_band_structure_and_dos().show()
    phonon.save()



import click
from clims.structure import Structure
from ase.io.formats import read, ioformats
from clims.cli.citations import cite_and_print


@click.command()
@click.option(
    "--filein",
    default="geometry.in",
    type=click.Path(exists=True),
    show_default=True,
    help="Name of the input file",
)
@click.option(
    "--format",
    default="aims",
    show_default=True,
    help="In- and output format of structure files",
)
@click.option("--symthresh", default=1e-5, help="Threshold for crystal symmetry detection")
@click.option("--primitive", is_flag=True, help="Write primitive to file.")
@click.option("--conventional", is_flag=True, help="Write conventional to file.")
def unit_cell_info(filein, format, symthresh, primitive, conventional):
    """
    Show information for the input structure. Mainly intended for periodic structures. The following quantities are written:

    * Number of atoms in the input structure.

    * Chemical formula of the input structure (alphabetically ordered).

    * The unit cell parameters (lattice constants and angles) of the input structure.

    * Name of the Bravais lattice of the refined (= symmetrized + standardized) structure (from spglib).

    * The spacegroup number and Hall symbol of the refined structure.

    * The Wyckoff positions of the refined structure (if spacegroup number > 1). The numbers in parenthesis denote the indeces of the unique atoms that occupy this position.

    * Whether the refined structure is primitive.

    * The used symmetry threshold.

    The input structure can be converted to the standardized primitive and conventional cell (using spglib).
    """
    try:
        c = read(filein, format=format)
    except:
        click.echo(f'ERROR: Could not identify format of input file "{filein}"')
        click.echo('Please specify the format of input file by "--format".')
        click.echo("Available formats are:")
        ase_formats = list(ioformats.keys())
        for a, b, c in zip(ase_formats[::3], ase_formats[1::3], ase_formats[2::3]):
            print("{:<30}{:<30}{:<}".format(a, b, c))
        raise
    si = Structure(c, symthresh)
    if (conventional or primitive) and not all(c.pbc):
        click.echo(
            f'ERROR: "{filein}" is not a periodic structure. Cannot write conventional or primitive cell.'
        )
        return
    if conventional:
        click.echo("Writing conventional unit cell (refined)")
        si.write_refined()
    if primitive:
        click.echo("Writing primitive unit cell (refined)")
        si.write_primitive()
    click.echo(si)

    cite_and_print(["ase", "spglib"])



from pathlib import Path
import os


def write_submission_file(
    header, mpi_command, aims_executable, output_name, error_name, path="."
):
    run_command = get_run_command(mpi_command, aims_executable, output_name, error_name)
    header += f"\n\n{run_command}"

    submit_name = Path(path) / "submit.sh"
    if not submit_name.is_file():
        with open(submit_name, "w") as submit_file:
            submit_file.write(header)
        st = os.stat(submit_name)
        os.chmod(submit_name, st.st_mode | 0o111)
    else:
        print("Submission file 'submit.sh' already exists. I'm not overwriting it.")


def get_run_command(mpi_command, aims_executable, output_name, error_name):
    run_command = f"{mpi_command} {aims_executable} > {output_name} 2> {error_name}"
    return run_command



from ase.io import read, write
import click
from clims.cli.citations import cite_and_print


@click.command()
@click.argument("init_geo_name")
@click.argument("uninit_geo_name")
@click.option("--fileout", default="geometry-reinitialized.in", type=str)
@click.option("--scaled", is_flag=True, help="Write geometry.in with fractional coordinates.")
def reinitialize_geometry(init_geo_name, uninit_geo_name, fileout, scaled):
    """
    Reinitialize UNINIT_GEO with the same initial_charges and initial_moments as
    in INIT_GEO.
    """
    init_geo = read(init_geo_name, format="aims")
    uninit_geo = read(uninit_geo_name, format="aims")
    assert len(init_geo) == len(uninit_geo)
    initial_moments = init_geo.get_initial_magnetic_moments()
    initial_charges = init_geo.get_initial_charges()
    print("-- Re-initializing with the following initial_moments:")
    print(initial_moments)
    print("-- Re-initializing with the following initial_charges:")
    print(initial_charges)
    uninit_geo.set_initial_magnetic_moments(initial_moments)
    uninit_geo.set_initial_charges(initial_charges)
    write(fileout, uninit_geo, scaled=scaled)

    restart_geo_footer = """#
# What follows is the current estimated Hessian matrix constructed by the BFGS algorithm.
# This is NOT the true Hessian matrix of the system.
# If you do not want this information here, switch it off using the "hessian_to_restart_geometry" keyword.
#
"""
    restart_geo = ""
    with open(uninit_geo_name) as f:
        lines = f.readlines()
    for line in lines:
        if line.strip().startswith("trust_radius"):
            restart_geo += line
        elif line.strip().startswith("hessian_file"):
            restart_geo += line
    if restart_geo:
        print(f"Appending geometry restart info to {fileout}")
        with open(fileout, "a") as f:
            f.write(restart_geo_footer + restart_geo)

    cite_and_print(["ase"])



import numpy as np
from pathlib import Path
import re


files = [x for x in Path(".").iterdir() if not x.is_dir()]


def get_file_name(pattern, files):
    for file in files:
        filename = file.as_posix()
        if pattern.search(filename) and not "raw" in filename and not "no_soc" in filename:
            return file
    raise FileNotFoundError


def get_mu(filename):
    f = open(filename)
    for _ in range(3):
        line = f.readline()
        if "chemical potential" in line:
            return float(line.split()[-2])
    raise


def read_species_dos(species, max_spin_channel):
    dos_species = {}
    maxdos = 0.0
    energy_lim = [100000, -100000]

    spinstrs = [""]
    if max_spin_channel == 2:
        spinstrs = ["_spin_up", r"_spin_d(ow)?n"]
    for s in species:
        dos_species[s] = []
        for ss in spinstrs:
            pattern = re.compile(s + r"_l_proj_dos(\_tetrahedron)?" + ss)
            filename = get_file_name(pattern, files)
            mu = get_mu(filename)
            data = np.loadtxt(filename)
            maxdos = max(maxdos, data[:, 1].max())
            dos_species[s].append(data)
            energy_lim = [
                min(data[:, 0][0], energy_lim[0]),
                max(data[:, 0][-1], energy_lim[1]),
            ]
    dos_species["maxdos"] = maxdos
    dos_species["energy_lim"] = energy_lim

    return dos_species


def get_dos_groups(groups, dos_species, max_spin_channel):
    dos_groups = {}
    for g in groups:
        species_in_g = g.split()
        max_l = 0
        for sg in species_in_g:
            max_l = max(max_l, dos_species[sg][0].shape[1])
        max_l -= 1  # First column is x axis. We store separately.
        dos_groups[g] = [
            np.zeros((dos_species[sg][0].shape[0], max_l))
            for i_spin in range(max_spin_channel)
        ]
        dos_groups["x"] = dos_species[sg][0][:, 0]
        for sg in species_in_g:
            tmp_max_l = dos_species[sg][0].shape[1] - 1
            for i_spin in range(max_spin_channel):
                dos_groups[g][i_spin][:, :tmp_max_l] += dos_species[sg][i_spin][:, 1:]
    return dos_groups


def read_total_dos():
    pattern = re.compile("KS_DOS_total")
    filename = get_file_name(pattern, files)
    mu = get_mu(filename)
    data = np.loadtxt(filename)
    energy = data[:, 0]
    dos = data[:, 1:]
    maxdos = dos.max()
    energy_lim = [energy[0], energy[-1]]
    return energy, dos, maxdos, energy_lim



import numpy as np


def read_bands(band_segments, band_totlength, max_spin_channel, gw):
    band_data = {}

    prev_end = band_segments[0][0]
    distance = band_totlength / 30.0  # distance between line segments that do not coincide

    xpos = 0.0
    labels = [(0.0, band_segments[0][4])]

    prefix = ""
    if gw:
        prefix = "GW_"

    for iband, (start, end, length, npoint, startname, endname) in enumerate(band_segments):
        band_data[iband + 1] = []
        if any(start != prev_end):
            xpos += distance
            labels += [(xpos, startname)]

        xvals = xpos + np.linspace(0, length, npoint)
        xpos = xvals[-1]

        labels += [(xpos, endname)]

        prev_end = end
        prev_endname = endname

        for spin in range(max_spin_channel):
            data = np.loadtxt(prefix + "band%i%03i.out" % (spin + 1, iband + 1))
            idx = data[:, 0].astype(int)
            kvec = data[:, 1:4]
            band_occupations = data[:, 4::2]
            band_energies = data[:, 5::2]
            assert (npoint) == len(idx)
            band_data[iband + 1] += [
                {
                    "xvals": xvals,
                    "band_energies": band_energies,
                    "color": "kr"[spin],
                    "band_occupations": band_occupations,
                }
            ]
    shift = shift_band_data(band_data, max_spin_channel)

    return band_data, labels, shift


def shift_band_data(band_data, max_spin_channel):

    vbm = -1000000000
    for n, segment in band_data.items():
        for spin in range(max_spin_channel):
            occs = segment[spin]["band_occupations"]
            for k in range(len(occs)):
                occ_diff = occs[k][1:] - occs[k][:-1]
                fermi_inds = np.nonzero(occ_diff != 0)[0]
                # print(fermi_inds)
                if len(fermi_inds) == 1:
                    tmp_vbm = segment[spin]["band_energies"][k][fermi_inds]
                    if vbm < tmp_vbm:
                        vbm = tmp_vbm
                else:
                    print("Your system is probably a metal. Not shifting bands.")
                    return 0

    # Shift all bands
    for n in band_data.keys():
        for spin in range(max_spin_channel):
            band_data[n][spin]["band_energies"] -= vbm[0]

    return -vbm[0]



import click
import json
from clims.testsuite_tools import generate_testsuite, get_directory_info


@click.group()
def testsuite():
    """
    Create a set of folders based on combinations of different 'control.in' settings.
    """
    pass

@testsuite.command()
@click.argument("calculator_json",type=click.File())
def create(calculator_json):
    """Create all possible combinations from the CALCULATOR_JSON file."""

    calc_set = json.load(calculator_json)
    generate_testsuite(calc_set)


@testsuite.command()
def example():
    """Get example JSON file containing the calculator"""

    calc_set = {
        "xc": ["pbe", "pz-lda"],
        "k_grid": [(2, 2, 2), (3, 3, 3)],
        "aims_command": ["orterun aims_1.x", "orterun aims_2.x"],
        "outfilename": "aims.out",
        "species_dir": [
            "/Users/sebastian/software/FHIaims/species_defaults/light",
            "/Users/sebastian/software/FHIaims/species_defaults/tight",
        ],
    }
    with open('calculator.json','w') as calculator_json:
        json.dump(calc_set, calculator_json, indent=4)


@testsuite.command()
def info():
    get_directory_info()



from click.testing import CliRunner
from clims.cli.supercell import supercell
from ase.build import bulk
from ase.io import write, read
import numpy as np


GaAs = bulk("GaAs", crystalstructure="zincblende", a=5.65315)
GaAs_positions_222 = [
    [0.0, 0.0, 0.0],
    [1.4132875, 1.4132875, 1.4132875],
    [2.826575, 2.826575, 0.0],
    [4.2398625, 4.2398625, 1.4132875],
    [2.826575, 0.0, 2.826575],
    [4.2398625, 1.4132875, 4.2398625],
    [5.65315, 2.826575, 2.826575],
    [7.0664375, 4.2398625, 4.2398625],
    [0.0, 2.826575, 2.826575],
    [1.4132875, 4.2398625, 4.2398625],
    [2.826575, 5.65315, 2.826575],
    [4.2398625, 7.0664375, 4.2398625],
    [2.826575, 2.826575, 5.65315],
    [4.2398625, 4.2398625, 7.0664375],
    [5.65315, 5.65315, 5.65315],
    [7.0664375, 7.0664375, 7.0664375],
]
GaAs_cell_222 = [
    [0.0, 5.65315, 5.65315],
    [5.65315, 0.0, 5.65315],
    [5.65315, 5.65315, 0.0],
]
GaAs_positions_cubic = [
    [0.0, 0.0, 0.0],
    [1.4132875, 1.4132875, 1.4132875],
    [2.826575, 2.826575, 0.0],
    [4.2398625, 4.2398625, 1.4132875],
    [2.826575, 0.0, 2.826575],
    [4.2398625, 1.4132875, 4.2398625],
    [0.0, 2.826575, 2.826575],
    [1.4132875, 4.2398625, 4.2398625],
]

clims_supercell_output = """
Read structure from:               : geometry.in
Chemical formula                   : AsGa
Number of atoms in original cell   : 2
Number of atoms in supercell       : 16
Supercell written to file          : geometry-supercell.in


Please consider to cite the following publications:

[ase]       Larsen, Ask Hjorth, et al.
            The Atomic Simulation Environment—A Python library for working with atoms
            J. Phys.: Condens. Matter 29, 273002 (2017)
            https://doi.org/10.1088/1361-648X/aa680e

"""

clims_supercell_format_error = """
***ERROR***: Could not identify format of input file "geometry.out"
-- Please specify the format of input file by "--format".
-- For available formats c.f.: https://wiki.fysik.dtu.dk/ase/ase/io/io.html
"""


def test_supercell_3_default():
    runner = CliRunner()
    with runner.isolated_filesystem():
        write("geometry.in", GaAs, scaled=True)
        result = runner.invoke(supercell, ["2", "2", "2"])
        assert result.exit_code == 0
        sc = read("geometry-supercell.in")
        np.testing.assert_allclose(sc.positions, GaAs_positions_222, atol=1e-12)
        assert result.output == clims_supercell_output


def test_supercell_9_default():
    runner = CliRunner()
    with runner.isolated_filesystem():
        write("geometry.in", GaAs, scaled=True)
        result = runner.invoke(supercell, ["2", "0", "0", "0", "2", "0", "0", "0", "2"])
        assert result.exit_code == 0
        sc = read("geometry-supercell.in")
        np.testing.assert_allclose(sc.positions, GaAs_positions_222, atol=1e-12)
        assert result.output == clims_supercell_output


def test_supercell_to_cubic():
    runner = CliRunner()
    with runner.isolated_filesystem():
        write("geometry.in", GaAs, scaled=True)
        result = runner.invoke(supercell, ["--to-cubic-with-natoms", "8"])
        assert result.exit_code == 0
        sc = read("geometry-supercell.in")
        np.testing.assert_allclose(sc.positions, GaAs_positions_cubic, atol=1e-12)


def test_supercell_format_error():
    runner = CliRunner()
    with runner.isolated_filesystem():
        write("geometry.out", GaAs, scaled=True, format="aims")
        result = runner.invoke(supercell, ["--filein", "geometry.out", "2", "2", "2"])
        assert result.exit_code == 0
        assert result.output == clims_supercell_format_error



from pathlib import Path
from click.testing import CliRunner
from ase.io import read
from ase.build import bulk
from test.file_comparison import compare_geo_files
from clims.cli.elastic_constants import clims_elastic


parent = Path(__file__).parent

elastic_constants_output = """Cij solution
------------------------------
 Solution rank:  3
 Square of residuals: 4.5e-05
 Relative singular values:
 1.0000   0.7071   0.6354  

Elastic tensor (GPa):
   C_11     C_12     C_44  
------------------------------
 277.49    85.71   143.26  

Please consider to cite the following publications:

[ase]       Larsen, Ask Hjorth, et al.
            The Atomic Simulation Environment—A Python library for working with atoms
            J. Phys.: Condens. Matter 29, 273002 (2017)
            https://doi.org/10.1088/1361-648X/aa680e

[elastic]   Jochym, Pawel T.
            Module for calculating elastic tensor of crystals
            software, https://github.com/jochym/Elastic
            https://doi.org/10.5281/zenodo.593721
"""


def test_elastic_constants():
    runner = CliRunner()
    with runner.isolated_filesystem():
        print(f"{parent / 'elastic'}")
        c = read(parent / "elastic" / "geometry.in")
        c.write("geometry.in")
        result = runner.invoke(clims_elastic, "initialize")
        assert result.exit_code == 0
        dirlist = sorted([d for d in Path(".").iterdir() if d.is_dir()])
        rlist = sorted([d for d in (parent / "elastic").iterdir() if d.is_dir()])
        print(dirlist)
        assert len(rlist) == len(dirlist)
        for d, r in zip(dirlist, rlist):
            compare_geo_files(d / "geometry.in", r / "geometry.in")
        result = runner.invoke(
            clims_elastic, ["postprocess", "--pwd", f"{parent / 'elastic'}"]
        )
        assert result.exit_code == 0
        assert "\n".join(result.output.splitlines()[-24:]) == elastic_constants_output



from clims.prepare_submission import write_submission_file
from clims.cli.configure import get_config_value
from ase.calculators.aims import Aims
from ase.io import read, write
from pathlib import Path
import itertools
import json


def generate_testsuite(calc_set):
    keys = []
    values = []

    for key, value in calc_set.items():
        if isinstance(value, list):
            length = len(value)
        else:
            length = 1
            value = [value]
        values.append(value)
        keys.append([key] * length)

    keys_product = itertools.product(*keys)
    values_product = itertools.product(*values)

    calcs = []
    for i in zip(keys_product, values_product):
        dict = {}
        for key, value in zip(*i):
            dict[key] = value
        calcs.append(dict)

    structure = read("geometry.in", format="aims")
    dir_list = []
    for i, c in enumerate(calcs):
        calc_dir = f"{i:03}"
        calc = Aims(**c)
        calc.directory = calc_dir
        calc.write_input(structure)
        with open(calc_dir + "/calculator.json", "w") as calc_file:
            json.dump(c, calc_file)
        submission_params = get_submission_params(c)
        write_submission_file(*submission_params, path=calc_dir)
        dir_list.append(calc_dir)
    write_submit_all(dir_list)

def write_submit_all(dir_list):
    dir_list_str = " ".join(dir_list)
    submit_command = get_config_value('submit_command')
    if "./" == submit_command:
        space = ''
    else:
        space = ' '
    print(dir_list_str)
    submit_all_str = '\n'.join((
        f'for i in  {dir_list_str}',
        'do',
        '    cd $i',
        f'    {submit_command}{space}submit.sh',
        '    cd ..',
        'done'
    ))
    with open('submit_all.sh','w') as submit_all:
        submit_all.write(submit_all_str)


def get_submission_params(c):
    if "outfilename" in c:
        output_name = c["outfilename"]
    else:
        output_name = get_config_value("output_name")

    if "aims_command" in c:
        command_split = c["aims_command"].split(" ")
        if len(command_split) == 2:
            mpi_command = command_split[0]
            aims_executable = command_split[1]
        elif len(command_split) == 1:
            mpi_command = ""
            aims_executable = command_split[0]
        else:
            print("Unexpected format for 'aims_command'.")
            print("Expected: <mpi_command> <aims_executable>")
            print("Or: <aims_executable>")
            return
    else:
        mpi_command = get_config_value("mpi_command")
        aims_executable = get_config_value("aims_executable")

    error_name = get_config_value("error_name")
    header = get_config_value("submission_header")

    return [header, mpi_command, aims_executable, output_name, error_name]


def get_directory_info():
    p = Path(".")
    dir_list = [x for x in p.iterdir() if x.is_dir()]

    calcs = {}
    for x in dir_list:
        calc_json = x / "calculator.json"
        if calc_json.exists():
            with open(calc_json, "r") as calc_file:
                c = json.load(calc_file)
            calcs[x.as_posix()] = c

    sorted_calcs = {
        k: v for k, v in sorted(calcs.items(), key=lambda item: int(item[0]))
    }

    for k, v in sorted_calcs.items():
        s = ""
        for ck, cv in v.items():
            s += f"{ck:8}: {str(cv):8}, "
        print(k, ":", s)



from setuptools import setup, find_packages
import sys


python_min_version = (3, 9)
python_requires = ">=" + ".".join(str(num) for num in python_min_version)


if sys.version_info < python_min_version:
    raise SystemExit("Python 3.9 or later is required!")

with open("README.md") as f:
    long_description = f.read()

setup(
    name="clims",
    description="Command-Line Interface for Materials Simulations",
    author="FHI-aims developers",
    url="https://gitlab.com/FHI-aims-club/utilities/clims",
    packages=find_packages(),
    install_requires=[
        "ase>=3.23",
        "spglib>=2.5.0",
        "phonopy>=2.9.3",
        "matplotlib>=3.4.1",
        "click>=7.1.2",
        "pytest>=6.2.3",
        "numpy<2.0",
        "Elastic>=0.1",
    ],
    extras_require={"mace": ["mace-torch==0.3.6"]},
    entry_points={
        "console_scripts": [
            "clims-unit-cell-info=clims.cli.unit_cell_info:unit_cell_info",
            "clims-supercell=clims.cli.supercell:supercell",
            "clims-prepare-run=clims.cli.prepare_run:prepare_run",
            "clims-configure=clims.cli.configure:clims_configure",
            "clims-testsuite=clims.cli.testsuite:testsuite",
            "clims-xyplot=clims.cli.xyplot:xyplot",
            "clims-convert-geometry=clims.cli.convert:convert",
            "clims-phonopy=clims.cli.phonons:clims_phonopy",
            "clims-aimsplot=clims.cli.aimsplot:aimsplot",
            "clims-wigner-seitz-cluster=clims.cli.wigner_seitz_cluster:construct",
            "clims-reinitialize-geometry=clims.cli.reinitialize_geometry:reinitialize_geometry",
            "clims-elastic=clims.cli.elastic_constants:clims_elastic",
            "clims-prerelax=clims.cli.mace:clims_prerelax",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: "
        "GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    version="0.7.1",
    license="LGPLv2.1+",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={"clims.test": ["*_default"]},
    python_requires=python_requires,
)



from ase.io import aims
from ase.data import atomic_numbers
import re
import numpy as np


def write_control(
    c,
    write_path,
    species_dir,
    relax=False,
    bands=False,
    bands_mulliken=False,
    dos=False,
    hse06=False,
    k_grid_density=6,
    **calc_kwargs,
):
    """
    Checks and writes control.in template.
    """
    if hse06:
        xc = {"xc": "hse06 0.11", "hse_unit": "bohr", "hybrid_xc_coeff": 0.25}
        if bands:
            xc["exx_band_structure_version"] = "1"
        calc_kwargs.update(**xc)
    elif not "xc" in calc_kwargs:
        calc_kwargs.update({"xc": "pbe"})

    if not "relativistic" in calc_kwargs:
        calc_kwargs.update({"relativistic": "atomic_zora scalar"})

    is_periodic = all(c.pbc)
    needs_spin = np.any(c.get_initial_magnetic_moments())

    output = []
    if is_periodic:
        r_lattice = np.linalg.norm(2.0 * np.pi * c.cell.reciprocal(), axis=1)
        if not "k_grid" in calc_kwargs or "k_grid_density" in calc_kwargs:
            k_grid = tuple(np.ceil(r_lattice * k_grid_density).astype(int))
            calc_kwargs["k_grid"] = k_grid
    if relax:
        calc_kwargs["relax_geometry"] = "trm 5e-3"
        if is_periodic:
            calc_kwargs["relax_unit_cell"] = "full"
    if dos:
        output += [
            "dos -20 10 15001 0.05",
            "species_proj_dos -20 10 15001 0.05",
        ]
        if is_periodic:
            calc_kwargs["dos_kgrid_factors"] = "3 3 3"
    if bands or bands_mulliken and is_periodic:
        print(f"-- Bravais Lattice: {c.cell.get_bravais_lattice().longname}")
        print(f"  {c.cell.get_bravais_lattice().description()}")
        bands = prepare_bandinput(c.get_cell(), bands_mulliken=bands_mulliken)
        output += bands
    if output:
        calc_kwargs["output"] = output
    if needs_spin:
        calc_kwargs["spin"] = "collinear"

    c_file = write_path / "control.in"

    if c_file.exists():
        print("-- Found control.in file")
        needs_xc, needs_ks, needs_species_default = check_control(is_periodic, c_file)
        # print(needs_xc, needs_ks, needs_species_default)
        if needs_species_default:
            print("-- Found control.in, but no species: Only attaching species")
            write_species(c, c_file, species_dir)
        else:
            print("-- Not attaching species_default (found some)")
        if needs_xc:
            print("-- Warning: Found control.in, but no xc flag")
        if needs_ks:
            print("-- Warning: Found control.in, but no k_grid or k_grid_density flag")
    else:
        print("-- Creating new control.in file.")
        calc_kwargs["species_dir"] = species_dir
        aims.write_control(c_file, c, calc_kwargs)


def prepare_bandinput(cell, density=35, bands_mulliken=False):
    """
    Prepares the band information needed for the FHI-aims control.in file.

    Parameters:

    max_points_per_path: int
        Number of kpoints per band path
    density: int
        Number of kpoints per Angstrom. Default: 35
    """
    from ase.dft.kpoints import resolve_kpt_path_string, kpoint_convert

    bp = cell.bandpath()
    # print(cell.get_bravais_lattice())
    r_kpts = resolve_kpt_path_string(bp.path, bp.special_points)
    band_string = "band"
    if bands_mulliken:
        band_string = "band_mulliken"
    linesAndLabels = []
    for labels, coords in zip(*r_kpts):
        dists = coords[1:] - coords[:-1]
        lengths = [np.linalg.norm(d) for d in kpoint_convert(cell, skpts_kc=dists)]
        points = np.int_(np.round(np.asarray(lengths) * density))
        # I store it here for now. Might be needed to get global info.
        linesAndLabels.append([points, labels[:-1], labels[1:], coords[:-1], coords[1:]])

    bands = []
    for segs in linesAndLabels:
        for points, lstart, lend, start, end in zip(*segs):
            points = max(points, 2)
            bands.append(
                "{} {:9.5f}{:9.5f}{:9.5f} {:9.5f}{:9.5f}{:9.5f} {:4} {:3}{:3}".format(
                    band_string, *start, *end, points, lstart, lend
                )
            )

    # print(bands)
    return bands


def check_control(is_periodic, control_filename):

    with open(control_filename, "r") as fd:
        control_file = fd.read()
    needs_xc = not bool(re.search(r"\s*xc\s+.+", control_file))
    needs_ks = (
        not bool(re.search(r"\s*(k_grid|k_grid_density)\s+.+\n", control_file)) and is_periodic
    )
    needs_species_default = not bool(re.search(r"\#  FHI\-aims code project", control_file))

    return needs_xc, needs_ks, needs_species_default


def write_species(c, c_file, species_dir, fallback_species_dir="light"):
    warning_indent = "  ***"
    symbols = [
        c.get_chemical_symbols()[i]
        for i in sorted(np.unique(c.get_chemical_symbols(), return_index=True)[1])
    ]
    control = open(c_file, "a")
    for symbol in symbols:
        fd = species_dir / f"{atomic_numbers[symbol]:02}_{symbol}_default"
        try:
            for line in open(fd, "r"):
                control.write(line)
        except:
            print(
                f"{warning_indent}Could not find {symbol} for {species_dir.name}. Trying: {fallback_species_dir}."
            )
            fd = (
                species_dir.parent
                / fallback_species_dir
                / f"{atomic_numbers[symbol]:02}_{symbol}_default"
            )
            try:
                for line in open(fd, "r"):
                    control.write(line)
            except:
                print(
                    f"{warning_indent}Could not find a species default for {symbol}. Stopping."
                )



import click
import numpy as np
from ase.io import read, write
from ase.build.supercells import make_supercell
from ase.build import find_optimal_cell_shape
from clims.cli.citations import cite_and_print


@click.command()
@click.argument("super_matrix", nargs=-1)
@click.option(
    "--filein",
    type=click.Path(exists=True),
    default="geometry.in",
    show_default=True,
    help="File of input structure",
)
@click.option("--format", type=str, default=None)
@click.option(
    "--scaled",
    type=bool,
    default=True,
    help="If output in sclaed coodinates",
    show_default=True,
)
@click.option(
    "--fileout",
    default="geometry-supercell.in",
    show_default=True,
    help="File name for supercell structure",
)
@click.option(
    "--to-cubic-with-natoms", type=int, help="Convert cell to most cubic with N_ATOMS."
)
def supercell(super_matrix, filein, fileout, format, scaled, to_cubic_with_natoms):
    """Create supercell from SUPER_MATRIX (3 or 9 elements)"""
    try:
        prim = read(filein, format=format)
    except StopIteration:
        click.echo(f'\n***ERROR***: Could not identify format of input file "{filein}"')
        click.echo('-- Please specify the format of input file by "--format".')
        click.echo(
            "-- For available formats c.f.: https://wiki.fysik.dtu.dk/ase/ase/io/io.html"
        )
        return

    if not all(prim.get_pbc()):
        click.echo("\nLoaded structure is not periodic.")
        return

    click.echo(f"\n{'Read structure from:':35}: {filein}")
    click.echo(f"{'Chemical formula':35}: {prim.get_chemical_formula()}")

    if to_cubic_with_natoms:
        click.echo(f"\nTry to find most cubic supercell with {to_cubic_with_natoms} atoms.")
        click.echo("(This may take some time ...)")
        to_cubic_with_natoms /= len(prim)  # ASE looks for number of unit cells.
        P = find_optimal_cell_shape(prim.cell, to_cubic_with_natoms, "sc", verbose=True)
    elif len(super_matrix) == 3:
        P = np.diag([int(i) for i in super_matrix])
    elif len(super_matrix) == 9:
        P = np.array([int(i) for i in super_matrix], dtype=np.double).reshape(3, 3)
    else:
        click.echo("The size of SUPER_MATRIX should be either 3 or 9.")
        return

    superatom = make_supercell(prim, P)
    click.echo(f"{'Number of atoms in original cell':35}: {len(prim)}")
    click.echo(f"{'Number of atoms in supercell':35}: {len(superatom)}")
    write(fileout, superatom, format=format, scaled=scaled)
    click.echo(f"{'Supercell written to file':35}: {fileout}\n")

    cite_and_print(["ase"])
